import { ANTLRInputStream, CommonTokenStream, Token } from 'antlr4ts';
import { ParseTreeWalker } from 'antlr4ts/tree/ParseTreeWalker'
import { UQM2ConvoScriptLexer } from './convoscript/UQM2ConvoScriptLexer';
import { UQM2ConvoScriptListener } from './convoscript/UQM2ConvoScriptListener';
import { I_addContext, I_conditionalContext, I_decContext, I_displayContext, I_gotoContext, I_incContext, I_removeContext, SegmentContext, StatesContext, UQM2ConvoScriptParser } from './convoscript/UQM2ConvoScriptParser';
import { Choice, ConvoScriptAdd, ConvoScriptConditional, ConvoScriptDec, ConvoScriptDisplay, ConvoScriptGoto, ConvoScriptInc, ConvoScriptRemove, ConvoStatement, Decision, Item, State, Statement } from './interfaces';

/**
 * @param input file contents to parse
 */
export function parseConvoScript(input: string) {
  function getTree(input: string) {
    let inputStream = new ANTLRInputStream(input);
    let lexer = new UQM2ConvoScriptLexer(inputStream);
    let tokenStream = new CommonTokenStream(lexer);
    let parser = new UQM2ConvoScriptParser(tokenStream);
    return parser.convo();
  }

  class Listener implements UQM2ConvoScriptListener {
    public choices: Array<Choice> = [];
    public decisions: Array<Decision> = [];
    public statements: Array<Statement> = [];
    public states: Array<State> = [];
    public items: Array<Item> = []; // TODO: support this.

    private cur_parsing: string | undefined = undefined;
    private cur_block: ConvoStatement[][] = [];

    enterChoices() {
      this.cur_parsing = "CHOICES";
    }

    enterDecisions() {
      this.cur_parsing = "DECISIONS";
    }

    enterStatements() {
      this.cur_parsing = "STATEMENTS";
    }

    enterStates() {
      this.cur_parsing = "STATES";
    }

    enterItems() {
      this.cur_parsing = "ITEMS";
    }

    exitSegment() {
      this.cur_block.pop();
    }

    enterSegment(context: SegmentContext) {
      if (this.cur_parsing === "CHOICES") {
        const choice: Choice = {
          id: context.header().UPPERCASE_WORD().text,
          playerStatementId: context.block()._statement.text,
          script: [] // to be filled in later!
        };
        this.cur_block.push(choice.script);
        this.choices.push(choice);
      } else if (this.cur_parsing === "DECISIONS") {
        const decision: Decision = {
          id: context.header().UPPERCASE_WORD().text,
          choiceIds: context.block().text().line().map(x => x.text.trim())
        };
        this.decisions.push(decision);
      } else if (this.cur_parsing === "STATEMENTS") {
        const statement: Statement = {
          id: context.header().STATEMENT_IDENTIFIER().text,
          text: context.block().text().line().map(x => x.text).join("\n")
        };
        this.statements.push(statement);
      } else if (this.cur_parsing === "STATES") {
        const state: State = {
          id: context.header().UPPERCASE_WORD().text,
          value: parseInt(context.block().text().line().map(x => x.text).join(""))
        };
        this.states.push(state);
      } else if (this.cur_parsing === "ITEMS") {
        //TODO
      } 
    }

    enterI_display(context: I_displayContext) {
      const display: ConvoScriptDisplay = {
        type: "DISPLAY",
        statementId: context._statement.text.trim()
      }
      this.cur_block[this.cur_block.length - 1].push(display);
    }

    enterI_remove(context: I_removeContext) {
      const display: ConvoScriptRemove = {
        type: "REMOVE",
        decisionId: context._decision.text.trim(),
        choiceId: context._choice.text.trim()
      }
      this.cur_block[this.cur_block.length - 1].push(display);
    }

    enterI_add(context: I_addContext) {
      const display: ConvoScriptAdd = {
        type: "ADD",
        decisionId: context._decision.text.trim(),
        choiceId: context._choice.text.trim()
      }
      this.cur_block[this.cur_block.length - 1].push(display);
    }

    enterI_inc(context: I_incContext) {
      const display: ConvoScriptInc = {
        type: "INC",
        variable: context._variable.text.trim()
      }
      this.cur_block[this.cur_block.length - 1].push(display);
    }

    enterI_dec(context: I_decContext) {
      const display: ConvoScriptDec = {
        type: "DEC",
        variable: context._variable.text.trim()
      }
      this.cur_block[this.cur_block.length - 1].push(display);
    }

    enterI_goto(context: I_gotoContext) {
      const display: ConvoScriptGoto = {
        type: "GOTO",
        decisionId: context._decision.text.trim(),
      }
      this.cur_block[this.cur_block.length - 1].push(display);
    }

    enterI_conditional(context: I_conditionalContext) {
      const condition: ConvoScriptConditional = {
        type: "CONDITION",
        variable: context._variable.text.trim(),
        comparator: context._comparison.text.trim(),
        value: context._value.text.trim(),
        body: []
      }
      this.cur_block[this.cur_block.length - 1].push(condition);
      this.cur_block.push(condition.body)
    }

    exitI_conditional() {
      this.cur_block.pop();
    }

    returnValue() {
      return {
        statements: this.statements,
        choices: this.choices,
        decisions: this.decisions,
        states: this.states,
        items: this.items
      }
    }

  }

  const listener = new Listener();
  ParseTreeWalker.DEFAULT.walk(listener as UQM2ConvoScriptListener, getTree(input));
  return listener.returnValue();
}
